"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _assert = _interopRequireDefault(require("assert"));

var _joi = _interopRequireDefault(require("joi"));

var _flattenDeep = _interopRequireDefault(require("lodash/flattenDeep"));

var _pick = _interopRequireDefault(require("lodash/pick"));

var _NotAllowed = _interopRequireDefault(require("../errors/NotAllowed"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(source, excluded) { if (source == null) return {}; var target = _objectWithoutPropertiesLoose(source, excluded); var key, i; if (Object.getOwnPropertySymbols) { var sourceSymbolKeys = Object.getOwnPropertySymbols(source); for (i = 0; i < sourceSymbolKeys.length; i++) { key = sourceSymbolKeys[i]; if (excluded.indexOf(key) >= 0) continue; if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue; target[key] = source[key]; } } return target; }

function _objectWithoutPropertiesLoose(source, excluded) { if (source == null) return {}; var target = {}; var sourceKeys = Object.keys(source); var key, i; for (i = 0; i < sourceKeys.length; i++) { key = sourceKeys[i]; if (excluded.indexOf(key) >= 0) continue; target[key] = source[key]; } return target; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class PermissionsManager {
  static generateCRUD(resource) {
    return ['create', 'read', 'update', 'delete'].map(action => ({
      permission: `${resource}.${action}`,
      description: `"${action}" action for the "${resource}" resource`
    }));
  }

  constructor({
    extractor = subject => subject.permissions,
    aliases = {
      view: 'read',
      access: 'read'
    }
  }) {
    this.permissions = {};
    this.aliases = aliases;
    this.extractor = extractor;
  }

  createAlias(forPermission, toPermission) {
    this.aliases[forPermission] = toPermission;
  }

  getAll() {
    return Object.keys(this.permissions).reduce((acc, permission) => [...acc, _objectSpread({
      permission
    }, (0, _pick.default)(this.permissions[permission], ['description', 'dependencies']))], []);
  }

  has(name) {
    return Object.keys(this.permissions).includes(name);
  }

  define(permission, rawConfig = {}) {
    if (Array.isArray(permission)) {
      return permission.forEach((_ref) => {
        let {
          permission: permissionName
        } = _ref,
            itemConfig = _objectWithoutProperties(_ref, ["permission"]);

        return this.define(permissionName, _objectSpread({}, rawConfig, itemConfig));
      });
    }

    const config = _joi.default.attempt(rawConfig, {
      description: _joi.default.string().default(`${permission} permission definition`),
      dependencies: _joi.default.array().default([]),
      check: _joi.default.func().default(() => true),
      isValidObject: _joi.default.func().default(() => true)
    });

    if (this.has(permission)) {
      throw new Error(`Can't add permission with a name of ${permission}. Already exists!`);
    }

    const dependencies = (0, _flattenDeep.default)(config.dependencies);
    dependencies.forEach(dependency => {
      if (!this.has(dependency)) {
        throw new Error(`Unable to find permission named "${dependency}" required as a dependency of "${permission}".`);
      }
    });
    this.permissions[permission] = config;
    return this;
  }

  async can(subject, permissionName, object) {
    try {
      await this.check(subject, permissionName, object);
    } catch (err) {
      if (err instanceof _NotAllowed.default) {
        return false;
      }

      throw err;
    }

    return true;
  }

  async checkAll(subject, permissions, object, parallel = true) {
    const walkPermissions = permissions.map(item => {
      if (Array.isArray(item)) {
        return () => this.checkAll(subject, item, object, !parallel);
      }

      return () => this.check(subject, item, object);
    });
    return parallel ? Promise.all(walkPermissions.map(task => task())) : walkPermissions.reduce((promise, task) => promise.then(task), Promise.resolve());
  }

  async check(subject, permissionName, object) {
    (0, _assert.default)(subject, 'Subject is required!');
    (0, _assert.default)(permissionName && typeof permissionName === 'string', 'Permission is required!');
    const name = Object.keys(this.aliases).includes(permissionName) ? this.aliases[permissionName] : permissionName;

    if (!this.has(name)) {
      throw new Error(`Unable to find permission named "${name}" when checking access.`);
    }

    const permission = this.permissions[name];
    const subjectPermissions = await this.extractor(subject);

    if (!subjectPermissions.includes(name)) {
      throw new _NotAllowed.default(`Failed permission check "${name}"! Reason: permission not included in the list of subjects permissions (${subjectPermissions}).`);
    }

    if (!permission.isValidObject(object)) {
      throw new _NotAllowed.default(`Failed permission check "${name}"! Reason: received invalid object (${object}).`);
    }

    await this.checkAll(subject, permission.dependencies, object);
    const isAllowed = await permission.check(subject, object);

    if (!isAllowed) {
      throw new _NotAllowed.default(`Failed permission check "${name}"! Reason: failed check function.`);
    }
  }

}

var _default = PermissionsManager;
exports.default = _default;